import time
import unittest
import PIL.Image
import torch
import torchvision
from irisml.tasks.benchmark_model import Task
from utils import FakeDataset


class TestBenchmarkModel(unittest.TestCase):
    def test_simple_model(self):
        class FakeModel(torch.nn.Module):
            def training_step(self, inputs, targets):
                time.sleep(0.01)
                return {'loss': torch.tensor(1.0, requires_grad=True)}

            def prediction_step(self, inputs):
                time.sleep(0.01)
                return

        model = FakeModel()
        dataset = FakeDataset([(PIL.Image.new('RGB', (224, 224)), torch.tensor(1))] * 100)
        transform = torchvision.transforms.ToTensor()
        outputs = Task(Task.Config(1, 'cpu')).execute(Task.Inputs(model, dataset, transform))

        self.assertGreater(outputs.forward_time_per_iteration, 0.01)
        self.assertGreater(outputs.forward_backward_time_per_iteration, 0.01)
        self.assertGreater(outputs.prediction_time_per_iteration, 0.01)
        self.assertEqual(outputs.max_cuda_memory_in_mb, 0)  # On CPU

    def test_skip_training(self):
        class FakeModel(torch.nn.Module):  # Since this model doesn't have training_step(), it will fail if skip_training=False.
            def prediction_step(self, inputs):
                time.sleep(0.01)
                return

        model = FakeModel()
        dataset = FakeDataset([(PIL.Image.new('RGB', (224, 224)), torch.tensor(1))] * 100)
        transform = torchvision.transforms.ToTensor()
        outputs = Task(Task.Config(1, 'cpu', skip_training=True)).execute(Task.Inputs(model, dataset, transform))

        self.assertEqual(outputs.forward_time_per_iteration, 0)
        self.assertEqual(outputs.forward_backward_time_per_iteration, 0)
        self.assertGreater(outputs.prediction_time_per_iteration, 0.01)
        self.assertEqual(outputs.max_cuda_memory_in_mb, 0)  # On CPU

    def test_simple_model_with_tuple_inputs(self):
        class FakeModel(torch.nn.Module):
            def training_step(self, inputs, targets):
                time.sleep(0.01)
                return {'loss': torch.tensor(1.0, requires_grad=True)}

            def prediction_step(self, inputs):
                time.sleep(0.01)
                return

        model = FakeModel()
        dataset = FakeDataset([((PIL.Image.new('RGB', (224, 224)), torch.tensor([1, 2, 3])), torch.tensor(1))] * 100)

        def transform(inputs):
            return torchvision.transforms.functional.to_tensor(inputs[0]), inputs[1]
        outputs = Task(Task.Config(1, 'cpu')).execute(Task.Inputs(model, dataset, transform))

        self.assertGreater(outputs.forward_time_per_iteration, 0.01)

    def test_model_without_step_method(self):
        class FakeModel(torch.nn.Module):
            @property
            def criterion(self):
                return torch.nn.CrossEntropyLoss()

            def forward(self, x):
                time.sleep(0.01)
                return torch.zeros((x.shape[0], 8), requires_grad=True)

        model = FakeModel()
        dataset = FakeDataset([(PIL.Image.new('RGB', (224, 224)), torch.tensor(1))] * 100)
        transform = torchvision.transforms.ToTensor()
        outputs = Task(Task.Config(1, 'cpu')).execute(Task.Inputs(model, dataset, transform))
        self.assertGreater(outputs.forward_time_per_iteration, 0.01)



import copy
import dataclasses
import logging
import statistics
import time
import typing
import torch
import irisml.core
from irisml.tasks.train.build_dataloader import build_dataloader
from irisml.tasks.train.plugin_list import PluginList
from irisml.tasks.train.plugin_loader import load_plugin

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Benchmark a given model using a given dataset.

    If num_iterations > 1, the result of the frist forward/backward pass will be excluded from the average calculation.

    Trimmed mean is used to calculate the average time. The fastest and slowest 20% of the iterations are excluded.

    Config:
        batch_size (int)
        device (Optional['cpu' or 'cuda']): If not provided, cuda is selected when CUDA is available.
        num_iterations (int): The number of iterations to run.
        skip_training (bool): If true, training steps are skipped and forward_* and backward_* will be zero.
        skip_prediction (bool): If true, prediction steps are skipped and prediction_time_per_iteration will be zero.
        plugins (List[str]): A list of plugins to use.
    """
    VERSION = '0.1.8'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module
        dataset: torch.utils.data.Dataset
        transform: typing.Callable

    @dataclasses.dataclass
    class Config:
        batch_size: int
        device: typing.Optional[typing.Literal['cpu', 'cuda']] = None
        num_iterations: int = 10
        skip_training: bool = False
        skip_prediction: bool = False
        plugins: typing.List[str] = dataclasses.field(default_factory=list)

    @dataclasses.dataclass
    class Outputs:
        forward_backward_time_per_iteration: float = 0.0
        forward_time_per_iteration: float = 0.0
        backward_time_per_iteration: float = 0.0
        prediction_time_per_iteration: float = 0.0
        max_cuda_memory_in_mb: float = 0.0

    def execute(self, inputs):
        plugins = PluginList([load_plugin(p, self.context) for p in self.config.plugins])
        device = self._get_device()
        self.device = device  # TODO: Better to make a trainer object.
        dataloader = build_dataloader(inputs.dataset, inputs.transform, batch_size=self.config.batch_size, shuffle=False, drop_last=False)
        forward_time_all = []
        backward_time_all = []
        prediction_time_all = []
        cuda_memory_at_start = 0

        model = copy.deepcopy(inputs.model)
        if device.type == 'cuda':
            torch.cuda.reset_peak_memory_stats(device)
            cuda_memory_at_start = torch.cuda.memory_allocated(device)

        model.to(device)

        if not self.config.skip_training:
            forward_time_all, backward_time_all = self._train(model, dataloader, device, plugins)

        if not self.config.skip_prediction:
            prediction_time_all = self._predict(model, dataloader, device, plugins)

        return self._get_outputs(forward_time_all, backward_time_all, prediction_time_all, device, cuda_memory_at_start)

    def _train(self, model, dataloader, device, plugins):
        forward_time_all = []
        backward_time_all = []
        model.train()
        plugins.on_train_start(self, model)
        plugins.on_train_epoch_start(self, model, epoch_index=0)
        # Measure training forward and backward pass
        for i, batch in enumerate(dataloader):
            batch = plugins.on_train_batch_start(self, model, batch, i)

            inputs, targets = batch
            inputs = self._to_device(inputs, device)
            targets = self._to_device(targets, device)

            loss, forward_time, backward_time = self._training_step(model, inputs, targets, device, plugins)
            plugins.on_train_batch_end(self, model, loss, batch, i)
            forward_time_all.append(forward_time)
            backward_time_all.append(backward_time)
            if i >= self.config.num_iterations:
                break

        plugins.on_train_epoch_end(self, model, epoch_index=0)

        if len(forward_time_all) < self.config.num_iterations:
            logger.info(f"The dataset is smaller than expected. The actual number of iteration is {len(forward_time_all)}")

        plugins.on_train_end(self, model)
        return forward_time_all, backward_time_all

    def _predict(self, model, dataloader, device, plugins):
        prediction_time_all = []
        model.eval()
        plugins.on_prediction_start(self, model)
        # Measure prediction time.
        for i, batch in enumerate(dataloader):
            batch = plugins.on_prediction_batch_start(self, model, batch, i)
            inputs, _ = batch
            inputs = self._to_device(inputs, device)
            prediction_time_all.append(self._prediction_step(model, inputs, device, plugins))
            if i >= self.config.num_iterations:
                break

        plugins.on_prediction_end(self, model)
        return prediction_time_all

    def _get_device(self) -> torch.device:
        """Get a torch device based on the configuration. If not specified explicitly, it uses cuda if available."""
        if self.config.device:
            device_name = self.config.device
        else:
            device_name = 'cuda' if torch.cuda.is_available() else 'cpu'
            logger.info(f"Training device is selected automatically: {device_name}. To specify the device manually, please set Config.device.")

        return torch.device(device_name)

    def _to_device(self, data, device):
        if isinstance(data, list):
            return [self._to_device(d, device) for d in data]
        elif isinstance(data, tuple):
            return tuple(self._to_device(d, device) for d in data)
        elif hasattr(data, 'to'):
            return data.to(device, non_blocking=True)
        return data

    @staticmethod
    def _synchronize(device):
        if device.type == 'cuda':
            torch.cuda.synchronize()

    def _training_step(self, model, inputs, targets, device, plugins):
        with plugins.forward_context():
            self._synchronize(device)
            start = time.time()
            if hasattr(model, 'training_step'):
                loss = model.training_step(inputs, targets)['loss']
            else:
                loss = model.criterion(model(inputs), targets)

            self._synchronize(device)
            forward_time = time.time() - start

        loss = plugins.on_train_backward_start(self, model, loss)

        start = time.time()
        loss.backward()
        self._synchronize(device)
        backward_time = time.time() - start

        return loss, forward_time, backward_time

    @torch.no_grad()
    def _prediction_step(self, model, inputs, device, plugins):
        with plugins.forward_context():
            self._synchronize(device)
            start = time.time()
            if hasattr(model, 'prediction_step'):
                model.prediction_step(inputs)
            elif hasattr(model, 'predictor'):
                model.predictor(model(inputs))
            else:
                model(inputs)
            self._synchronize(device)
            return time.time() - start

    def _get_outputs(self, forward_time_all, backward_time_all, prediction_time_all, device, cuda_memory_at_start):
        forward_time_per_iteration = self._trimmed_mean_without_first_sample(forward_time_all)
        backward_time_per_iteration = self._trimmed_mean_without_first_sample(backward_time_all)
        forward_backward_time_per_iteration = forward_time_per_iteration + backward_time_per_iteration
        prediction_time_per_iteration = self._trimmed_mean_without_first_sample(prediction_time_all)
        max_cuda_memory_in_mb = ((torch.cuda.max_memory_allocated(device) - cuda_memory_at_start) / 2 ** 20) if device.type == 'cuda' else 0

        logger.debug(f"{forward_time_all=}")
        logger.debug(f"{backward_time_all=}")
        if prediction_time_all:
            logger.debug(f"{prediction_time_all=}")

        logger.info(f"{forward_time_per_iteration=}, {backward_time_per_iteration=}, {forward_backward_time_per_iteration=}, {prediction_time_per_iteration=}")
        if max_cuda_memory_in_mb > 0:
            logger.info(f"{max_cuda_memory_in_mb=}")

        return self.Outputs(forward_backward_time_per_iteration, forward_time_per_iteration, backward_time_per_iteration, prediction_time_per_iteration, max_cuda_memory_in_mb)

    @staticmethod
    def _trimmed_mean_without_first_sample(values):
        if not values:
            return 0

        # Since the first run requires initialization and weights transfer, it is usually slow. We exlucde it from calculating the average.
        values = values[1:] if len(values) > 2 else values

        # Trim the highest/lowest 20% of the values.
        k = int(len(values) * 0.2)
        values = sorted(values)[k:-k] if k > 0 else values

        return statistics.fmean(values)



import unittest
import torch
from irisml.tasks.remove_empty_images_from_dataset import Task

from utils import FakeDatasetWithGetTargets


class TestRemoveEmptyImagesFromDataset(unittest.TestCase):
    def test_object_detection(self):
        data = [(None, torch.tensor([[i, 0, 0, 1, 1]])) for i in range(10)]
        outputs = Task(Task.Config()).execute(Task.Inputs(FakeDatasetWithGetTargets(data)))
        self.assertEqual(len(outputs.dataset), 10)

        data = [(None, torch.tensor([[i, 0, 0, 1, 1]])) for i in range(10)] + [(None, torch.tensor([])) for _ in range(10)]
        outputs = Task(Task.Config()).execute(Task.Inputs(FakeDatasetWithGetTargets(data)))
        self.assertEqual(len(outputs.dataset), 10)

    def test_classification(self):
        data = [(None, torch.tensor([i])) for i in range(10)]
        outputs = Task(Task.Config()).execute(Task.Inputs(FakeDatasetWithGetTargets(data)))
        self.assertEqual(len(outputs.dataset), 10)

        data = [(None, torch.tensor([i])) for i in range(10)] + [(None, torch.tensor([])) for _ in range(10)]
        outputs = Task(Task.Config()).execute(Task.Inputs(FakeDatasetWithGetTargets(data)))
        self.assertEqual(len(outputs.dataset), 10)



"""Freeze parameters up to the specified module.

The order of modules are determined by running a model with forward hooks.
"""

import functools
import logging
import re
import torch
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, module_name_pattern):
        """
        Args:
            module_name_pattern (str): A regular expression to find a module. If empty, this plugin does nothing.
        """
        self._module_name_pattern = module_name_pattern

    def on_train_start(self, trainer, model):
        self._frozen = False

    def on_train_batch_start(self, trainer, model, batch, batch_index):
        if not self._frozen:
            freeze_targets = get_module_names_upto(model, self._module_name_pattern, batch[0], trainer.device)
            freeze_modules(model, freeze_targets)
            if hasattr(trainer, 'recreate_ddp_model'):
                trainer.recreate_ddp_model()  # DDP model needs to be created to exclude the frozen parameters.
            self._frozen = True

        return batch

    def on_train_end(self, trainer, model):
        model.requires_grad_(True)
        model.train()


def get_module_names_upto(module, module_name_pattern, example_inputs, device):
    """Get a list of module names up to the module that matches the given name pattern."""
    called_modules = []

    def log_module_forward(module, inputs, outputs, module_name):
        called_modules.append(module_name)

    # Add hooks
    hooks = [m.register_forward_hook(functools.partial(log_module_forward, module_name=n)) for n, m in module.named_modules()]

    logger.debug(f"Running the model once on {device} to get the order of modules.")
    example_inputs = to_device(example_inputs, device)
    with torch.no_grad():
        # Some module parameters might be impacted by this forward pass. Might need to restore the initial state if it caused an issue.
        module(example_inputs)

    for hook in hooks:
        hook.remove()

    assert called_modules
    pattern = re.compile(module_name_pattern)
    matched_index = next((i for i, n in enumerate(called_modules) if pattern.match(n)), None)
    if matched_index is None:
        logger.error(f"Failed to get matching module names. {module_name_pattern=}, {called_modules=}")
        return []
    else:
        return called_modules[:matched_index]


def freeze_modules(module, module_names):
    for n, m in module.named_modules():
        if n in module_names:
            logger.debug(f"Freezing module {n}")
            m.requires_grad_(False)
            m.eval()


def to_device(tensor_or_list, device):
    if hasattr(tensor_or_list, 'to'):
        return tensor_or_list.to(device)
    elif isinstance(tensor_or_list, list):
        return [to_device(x, device) for x in tensor_or_list]
    else:
        return tensor_or_list



"""Freeze parameters as per provided freeze pattern(s)."""

import logging
import re
import torch
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, freeze_pattern: str):
        """
        Args:
            freeze_pattern (str): A regular expression to find a module. If empty, this plugin does nothing.
        """

        self._freeze_pattern = freeze_pattern and re.compile(freeze_pattern)
        if not freeze_pattern:
            logger.info('No module name patterns provided. Skipping freezing modules.')

    def on_train_start(self, trainer, model):
        if self._freeze_pattern:
            _freeze(self._freeze_pattern, model)
            if hasattr(trainer, 'recreate_ddp_model'):
                trainer.recreate_ddp_model()  # DDP model needs to be created to exclude the frozen parameters.

    def on_train_end(self, trainer, model):
        model.requires_grad_(True)
        model.train()


def _freeze(freeze_pattern: re.Pattern, model: torch.nn.Module):
    for name, module in model.named_modules():
        if freeze_pattern.search(name):
            logger.debug(f"Freezing module {name}")
            module.requires_grad_(False)
            module.eval()



import unittest
import PIL.Image
import torch
from irisml.tasks.make_classification_dataset_from_object_detection import Task
from utils import FakeDataset


class TestMakeClassificationDatasetFromObjectDetection(unittest.TestCase):
    def test_simple(self):
        fake_image = PIL.Image.new('RGB', (32, 32))
        dataset = FakeDataset([(fake_image, torch.tensor([[0, 0, 0, 0.5, 0.5], [1, 0, 0, 0.1, 0.1]])),
                               (fake_image, torch.tensor([[2, 0, 0, 0.5, 0.5], [3, 0, 0, 0.1, 0.1]]))])

        outputs = Task(Task.Config()).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs.dataset), 4)
        self.assertEqual(outputs.dataset[0][0].size, (16, 16))
        self.assertEqual(outputs.dataset[0][1], 0)
        self.assertEqual(outputs.dataset[1][0].size, (3, 3))
        self.assertEqual(outputs.dataset[1][1], 1)
        self.assertEqual(outputs.dataset[2][0].size, (16, 16))
        self.assertEqual(outputs.dataset[2][1], 2)
        self.assertEqual(outputs.dataset[3][0].size, (3, 3))
        self.assertEqual(outputs.dataset[3][1], 3)

    def test_given_multiclass_dataset(self):
        fake_image = PIL.Image.new('RGB', (32, 32))
        dataset = FakeDataset([(fake_image, torch.tensor(0)),
                               (fake_image, torch.tensor(1))])
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs.dataset), 2)
        self.assertEqual(outputs.dataset[0][0].size, (32, 32))
        self.assertEqual(outputs.dataset[0][1], 0)
        self.assertEqual(outputs.dataset[1][1], 1)

    def test_given_multilabel_dataset(self):
        fake_image = PIL.Image.new('RGB', (32, 32))
        dataset = FakeDataset([(fake_image, torch.tensor([0, 1])),
                               (fake_image, torch.tensor([2, 3, 4]))])
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs.dataset), 2)
        self.assertEqual(outputs.dataset[0][0].size, (32, 32))
        self.assertEqual(outputs.dataset[0][1].dim(), 1)
        self.assertEqual(outputs.dataset[1][1][2], 4)



import unittest
import torch
from irisml.tasks.get_targets_from_dataset import Task

from utils import FakeDataset


class FakeDatasetWithGetTargets(FakeDataset):
    def __init__(self, data):
        self._data = data
        self._counter = 0

    @property
    def call_count(self):
        return self._counter

    def get_targets(self, index):
        self._counter += 1
        return self._data[index][1]

    def __len__(self):
        return len(self._data)

    def __getitem__(self, index):
        raise RuntimeError


class TestGetTargetsFromDataset(unittest.TestCase):
    def test_simple(self):
        data = [('image0', 0), ('image1', 2), ('image2', 4)]
        inputs = Task.Inputs(FakeDataset(data))
        outputs = Task(Task.Config()).execute(inputs)

        targets = outputs.targets

        self.assertIsInstance(targets, torch.Tensor)
        self.assertEqual(targets.tolist(), [0, 2, 4])

    def test_object_detection(self):
        data = [(None, torch.tensor([[i, 0, 0, 1, 1]])) for i in range(100)]
        outputs = Task(Task.Config()).execute(Task.Inputs(FakeDataset(data)))

        for i in range(100):
            self.assertTrue(torch.equal(outputs.targets[i], torch.tensor([[i, 0, 0, 1, 1]])))

    def test_get_targets(self):
        dataset = FakeDatasetWithGetTargets([(None, torch.tensor(i)) for i in range(100)])
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset))
        self.assertTrue(torch.equal(outputs.targets, torch.tensor([i for i in range(100)])))
        self.assertEqual(dataset.call_count, 100)



import unittest
import torch
from irisml.tasks.exclude_negative_samples_from_classification_dataset import Task


class TestExcludeNegativeSamplesFromClassificationDataset(unittest.TestCase):
    def test_multiclass(self):
        dataset = [(None, torch.tensor(0)), (None, torch.tensor(-1)), (None, torch.tensor(2)), (None, torch.tensor(-1))]
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset=dataset))
        self.assertEqual(len(outputs.dataset), 2)
        self.assertEqual(outputs.dataset[0][1], torch.tensor(0))
        self.assertEqual(outputs.dataset[1][1], torch.tensor(2))
        with self.assertRaises(IndexError):
            outputs.dataset[3]
        self.assertFalse(hasattr(outputs.dataset, 'get_targets'))



import unittest
import torch
from irisml.tasks.find_incorrect_classification_indices import Task


class TestFindIncorrectClassificationIndices(unittest.TestCase):
    def test_1d(self):
        predictions = torch.tensor([0, 1, 2, 2, 2, 5])
        targets = torch.tensor([0, 1, 2, 3, 4, 5])
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions, targets))
        self.assertEqual(outputs.indices.tolist(), [3, 4])

    def test_2d(self):
        predictions = torch.tensor([[0, 0, 0, 0, 0, 1],
                                    [0, 1, 0, 0, 0, 0],
                                    [0, 0, 0, 1, 0, 0],
                                    [0, 0, 0, 1, 0, 0],
                                    [0, 1, 0, 0, 0, 0],
                                    [0, 0, 0, 0, 0, 1]])
        targets = torch.tensor([0, 1, 2, 3, 4, 5])
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions, targets))
        self.assertEqual(outputs.indices.tolist(), [0, 2, 4])



import dataclasses
import logging
import typing
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Create a dataset with a list of strings.

    This dataset returns a tuple of ((text, []), None) for each item.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        texts: typing.List[str]

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        logger.info(f"Creating a fixed text dataset with {len(inputs.texts)} texts")
        dataset = FixedTextDataset(inputs.texts)
        return self.Outputs(dataset=dataset)

    def dry_run(self, inputs):
        return self.execute(inputs)


class FixedTextDataset(torch.utils.data.Dataset):
    def __init__(self, texts):
        super().__init__()
        self._texts = texts

    def __len__(self):
        return len(self._texts)

    def __getitem__(self, index):
        return (self._texts[index], []), None



import unittest
from irisml.tasks.num_iters_to_epochs import Task
from utils import FakeDataset


class TestNumItersToEpochs(unittest.TestCase):
    def test_more_than_1_epoch(self):
        task = Task(Task.Config())
        epochs = task.execute(Task.Inputs(batch_size=16, num_iterations=10, dataset=FakeDataset([1, 2, 3, 4]))).num_epochs
        self.assertEqual(epochs, 40)

    def test_less_than_1_epoch(self):
        task = Task(Task.Config())
        epochs = task.execute(Task.Inputs(batch_size=2, num_iterations=1, dataset=FakeDataset([1, 2, 3, 4]))).num_epochs
        self.assertEqual(epochs, 1)



import pathlib
import tempfile
import unittest
import PIL.Image
from irisml.tasks.load_simple_classification_dataset import Task


class TestLoadSimpleClassificationDataset(unittest.TestCase):
    def test_simple(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)
            index_filepath = temp_dir / 'images.txt'
            index_filepath.write_text('\n'.join(['image0.jpg 0', 'image1.jpg 1', 'image2.jpg 0']))
            labels_filepath = temp_dir / 'labels.txt'
            labels_filepath.write_text('\n'.join(['cat', 'dog']))
            PIL.Image.new('RGB', (10, 10)).save(temp_dir / 'image0.jpg')
            PIL.Image.new('RGB', (10, 10)).save(temp_dir / 'image1.jpg')
            PIL.Image.new('RGB', (10, 10)).save(temp_dir / 'image2.jpg')
            outputs = Task(Task.Config(index_filepath)).execute(Task.Inputs())

            self.assertEqual(outputs.num_classes, 2)
            self.assertEqual(outputs.class_names, ['cat', 'dog'])
            self.assertEqual(len(outputs.dataset), 3)
            self.assertEqual(outputs.dataset[0][1], 0)
            self.assertEqual(outputs.dataset[1][1], 1)
            self.assertEqual(outputs.dataset[2][1], 0)
            self.assertIsInstance(outputs.dataset[0][0], PIL.Image.Image)
            self.assertIsInstance(outputs.dataset[1][0], PIL.Image.Image)
            self.assertIsInstance(outputs.dataset[2][0], PIL.Image.Image)



"""Model Exponential Moving Average (EMA)

This plugin keeps expornential moving average of the model parameter as it is training.

The EMA weights are updated as follows at the end of each epoch:
    ema_weights = decay * ema_weights + (1 - decay) * current_weights

Parameters:
    decay (float): the decay parameter. See the comments above.
    num_skip_iterations (int): The number of iterations this plugin waits to initialize the averaged model.
    use_cpu_memory (bool): Use the cpu memory to keep the averaged model. Slow, but useful when there is not enough CUDA memory.
"""
import logging
import torch
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, decay: float = 0.99, num_skip_iterations: int = 100, use_cpu_memory: bool = False):
        if not (0.0 <= decay <= 1.0):
            raise ValueError(f"Decay must be in the range [0, 1]. decay={decay}")
        self._decay = decay
        self._num_skip_iterations = num_skip_iterations
        self._use_cpu_memory = use_cpu_memory
        self._averaged_model = None
        self._counter = 0

    def on_train_start(self, trainer, model):
        if not self._num_skip_iterations:
            self._init_ema(model)
        self._counter = 0
        self._weights_updated = False

    def on_optimizer_step_start(self, trainer, model, optimizer):
        self._weights_updated = True
        return optimizer

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        if self._weights_updated:  # Weights might not be updated when gradient accumulation was used.
            if self._averaged_model:
                self._averaged_model.update_parameters(model)
            elif self._counter >= self._num_skip_iterations - 1:
                self._init_ema(model)
            else:
                self._counter += 1
            self._weights_updated = False

    def on_train_end(self, trainer, model):
        if self._averaged_model:
            self._copy_weights(list(self._averaged_model.parameters()), list(model.parameters()))
            logger.info("Updated the model with the averaged parameters.")
        self._averaged_model = None

    def on_validation_start(self, trainer, model):
        if self._averaged_model:
            self._model_weights_backup = [p.detach().clone().to('cpu') for p in model.parameters()]
            self._copy_weights(list(self._averaged_model.parameters()), list(model.parameters()))
            logger.debug("Copying the averaged parameters to the model.")

    def on_validation_end(self, trainer, model):
        if self._averaged_model:
            self._copy_weights(self._model_weights_backup, list(model.parameters()))
            self._model_weights_backup = None

    def _init_ema(self, model):
        def ema_avg(averaged_model_params, model_params, num_averaged):
            return self._decay * averaged_model_params + (1 - self._decay) * model_params

        # Keep the copy on CPU memory.
        device = torch.device('cpu') if self._use_cpu_memory else next(model.parameters()).device
        self._averaged_model = torch.optim.swa_utils.AveragedModel(model, avg_fn=ema_avg, device=torch.device('cpu'))
        logger.info(f"Model EMA is enabled with decay={self._decay}, {device=}")

    @staticmethod
    def _copy_weights(src_params, dst_params):
        assert len(src_params) == len(dst_params)
        for src_param, dst_param in zip(src_params, dst_params):
            assert src_param.shape == dst_param.shape
            dst_param.detach().copy_(src_param.to(dst_param.device))



import copy
import dataclasses
import torch.nn
import irisml.core


class Task(irisml.core.TaskBase):
    """Split a image-text model into an image model and a text model.

    Inputs:
        model (torch.nn.Module): An input model. It must have 'image_model' and 'text_model' attributes.

    Config:
        append_normalizer (bool): If True, append a normalization module to the end of the image and text models.
    """
    VERSION = '0.1.2'

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module

    @dataclasses.dataclass
    class Config:
        append_normalizer: bool = True

    @dataclasses.dataclass
    class Outputs:
        image_model: torch.nn.Module
        text_model: torch.nn.Module
        logit_scale: torch.Tensor

    def execute(self, inputs):
        image_modules = [copy.deepcopy(inputs.model.image_model), copy.deepcopy(inputs.model.image_projection)]
        text_modules = [copy.deepcopy(inputs.model.text_model), copy.deepcopy(inputs.model.text_projection)]

        if self.config.append_normalizer:
            image_modules.append(NormModule())
            text_modules.append(NormModule())

        image_model = torch.nn.Sequential(*image_modules)
        text_model = torch.nn.Sequential(*text_modules)
        logit_scale = copy.deepcopy(inputs.model.logit_scale)
        return self.Outputs(image_model, text_model, logit_scale)

    def dry_run(self, inputs):
        return self.execute(inputs)


class NormModule(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        x = x / x.norm(dim=-1, keepdim=True)
        return x



import unittest
import torch
from irisml.tasks.benchmark_dataset import Task


class TestBenchmarkDataset(unittest.TestCase):
    def test_simple(self):
        dataset = torch.utils.data.TensorDataset(torch.randn(100, 3, 32, 32), torch.randint(0, 10, (100,)))

        def fake_transform(x):
            return x

        outputs = Task(Task.Config(batch_size=1)).execute(Task.Inputs(dataset=dataset, transform=fake_transform))
        self.assertGreater(outputs.time_per_batch, 0.0)



import dataclasses
import logging
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Calculate accuracy of the given prediction results.

    This task supports only multiclass classification.

    Inputs:
        predictions: The prediction results. The shape is (N, num_classes) or (N, ). If the shape is (N, ), the values are the indexes of the maximum values.
        targets: The ground truth labels. The shape is (N, ) or (N, 1).
    """
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Inputs:
        predictions: torch.Tensor  # Shape: (N, num_classes) or (N, )
        targets: torch.Tensor  # Shape: (N, ) or (N, 1)

    @dataclasses.dataclass
    class Outputs:
        accuracy: float = 0

    def execute(self, inputs):
        if len(inputs.predictions) == 0 or len(inputs.targets) == 0:
            raise ValueError("The predictions or the targets are empty.")

        if inputs.predictions.shape[0] != inputs.targets.shape[0]:
            raise ValueError(f"The predictions or the targets have unexpected shape: {inputs.predictions.shape} vs {inputs.targets.shape}")

        if len(inputs.predictions.shape) == 1:
            predicted_max_indexes = inputs.predictions
        elif len(inputs.predictions.shape) == 2:
            _, predicted_max_indexes = inputs.predictions.max(dim=1)
            predicted_max_indexes = predicted_max_indexes.flatten()
        else:
            raise ValueError(f"Unexpected shape of the predictions: {inputs.predictions.shape}")

        targets = inputs.targets.flatten()

        if targets.shape != predicted_max_indexes.shape:
            raise RuntimeError(f"The predictions or the targets have unexpected shape: {targets.shape} vs {predicted_max_indexes.shape}")

        accuracy = ((predicted_max_indexes == targets).sum() / len(targets)).item()

        logger.info(f"Accuracy: {accuracy:.3f}")
        return self.Outputs(accuracy)



"""Log the current status every N minutes. Useful for large datasets that an epoch takes long time.

DEBUG level is used for logging.
"""
import logging
import time
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, interval_in_minutes=5):
        self._interval_in_minutes = interval_in_minutes

    def on_train_epoch_start(self, trainer, model, epoch_index):
        self._last_update = time.time()
        return True

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        if time.time() > self._last_update + self._interval_in_minutes * 60:
            self._last_update = time.time()
            logger.debug(f"Batch {batch_index} is completed. loss={float(loss.item())}")

    def on_prediction_start(self, trainer, model):
        self._last_update = time.time()

    def on_prediction_batch_end(self, trainer, model, batch, batch_index):
        if time.time() > self._last_update + self._interval_in_minutes * 60:
            self._last_update = time.time()
            logger.debug(f"Batch {batch_index} is completed.")



import collections
import dataclasses
import logging
import math
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Make an oversampled dataset.

    Config:
        min_num_samples (int): The minimum number of samples in the oversampled dataset.
        oversampling_rate (float): The rate of oversampling. Actual rate can be bigger than this value because of the effect of min_num_samples.
        balance (bool): If True, oversample from the minority classes.
    """
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Config:
        min_num_samples: int
        oversampling_rate: float = 1.0
        balance: bool = False
        balance_weights_max: float = 5.0
        balance_weights_min: float = 0.2

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        num_samples = len(inputs.dataset)
        num_new_samples = max(0, self.config.min_num_samples - num_samples, int((self.config.oversampling_rate - 1) * num_samples))

        if num_new_samples > 0:
            logger.info(f"Oversampling the dataset. The current number of samples: {num_samples}. New: {num_samples+num_new_samples}")
            if self.config.balance:
                weights = self._calculcate_weights(inputs.dataset, self.config.balance_weights_max, self.config.balance_weights_min)
                new_samples = torch.multinomial(torch.tensor(weights), num_new_samples, replacement=True)
            else:
                new_samples = torch.randint(high=num_samples, size=(num_new_samples,), dtype=torch.int64)

            index_map = torch.cat([torch.arange(num_samples, dtype=torch.int64), new_samples]).tolist()
            dataset = OversampledDataset(inputs.dataset, index_map)
        else:  # No oversampling
            dataset = inputs.dataset
            if self.config.balance:
                logger.warning("Dataset balancing is not performed. Please change oversampling_rate to a higher value.")

        assert len(dataset) == num_samples + num_new_samples
        return self.Outputs(dataset)

    def dry_run(self, inputs):
        return self.Outputs(inputs.dataset)

    @staticmethod
    def _calculcate_weights(dataset, balance_weights_max, balance_weights_min):
        # Get the number of annotations per class.
        class_ids = []
        for i in range(len(dataset)):
            if hasattr(dataset, 'get_targets'):
                targets = dataset.get_targets(i)
            else:
                targets = dataset[i][1]
            class_ids.append(get_class_id(targets))
        class_counter = collections.Counter()
        for ids in class_ids:
            class_counter.update(ids)

        mean_num_tags = sum(class_counter.values()) / len(class_counter)
        weights_per_class = {c: mean_num_tags / n for c, n in class_counter.items()}

        # Make soft target
        weights_per_class = {c: math.sqrt(w) for c, w in weights_per_class.items()}

        # Clamp the weights
        weights_per_class = {c: max(min(w, balance_weights_max), balance_weights_min) for c, w in weights_per_class.items()}

        weights_per_sample = [math.prod(weights_per_class[t] for t in tags) for tags in class_ids]
        weights_per_sample = [min(w, balance_weights_max) for w in weights_per_sample]

        return weights_per_sample


def get_class_id(targets):
    """Get a list of class ids from the target.
    Supported target formats:
        LongTensor(,)
        LongTensor(N,)
        FloatTensor(N,)
        FloatTensor(N, 5)
    """
    if not targets.shape:
        return [int(targets)] if not targets.is_floating_point() else []
    elif len(targets.shape) == 1:
        return list(set(targets.tolist())) if not targets.is_floating_point() else []
    elif len(targets.shape) == 2:
        return list(set(targets[:, 0].flatten().tolist()))
    raise TypeError(f"Unsupported target type is detected: {targets} ({type(targets)})")


class OversampledDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, index_map):
        self._dataset = dataset
        self._index_map = index_map

    def __len__(self):
        return len(self._index_map)

    def __getitem__(self, index):
        new_index = self._index_map[index]
        assert isinstance(new_index, int)
        return self._dataset[new_index]



import dataclasses
import logging
import time
import torch
from irisml.tasks.train_with_gradient_cache import GradientCachingMixin, GradientCachingTrainer
from irisml.tasks.benchmark_model import Task as BenchmarkModelTask

logger = logging.getLogger(__name__)


class Task(BenchmarkModelTask):
    """Benchmark a given model using a given dataset with grad caching. Useful for cases which require sub batching.

    Config:
        sub_batch_size (int)
    """
    VERSION = '0.0.2'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Config(BenchmarkModelTask.Config):
        sub_batch_size: int = 1

    def _training_step(self, model, inputs, targets, device, plugins):
        assert self.config.batch_size % self.config.sub_batch_size == 0, 'batch_size must be divisible by sub_batch_size'
        grad_cache_wrapper = BenchmarkGradientCachingTrainer(self.config.sub_batch_size, plugins)
        return grad_cache_wrapper._training_step(model, inputs, targets, device)


class BenchmarkGradientCachingTrainer(GradientCachingTrainer):

    def __init__(self, sub_batch_size, plugins):
        self._sub_batch_size = sub_batch_size
        self._plugins = plugins

    def _training_step(self, model, inputs, targets, device):

        sub_batches = list(zip(self._split_tensor_or_list(inputs), self._split_tensor_or_list(targets)))
        _model = GradientCachingMixin.ModelWrapper(model, device)

        self._synchronize(device)
        start = time.time()
        _batch = (inputs, targets)
        loss, features, features_local, rng_states = self._forward_pass(_model, _batch, sub_batches)
        self._synchronize(device)
        forward_time_first = time.time() - start

        start = time.time()
        loss = self._calculate_gradients(model, loss)
        self._synchronize(device)
        # NOTE: not the actual backward pass, but the time it takes to calculate the gradients with a backward pass
        backward_time = time.time() - start

        start = time.time()
        self._forward_backward_pass_with_gradients(model, _model, features, features_local, rng_states, sub_batches)
        self._synchronize(device)
        forward_time_second = time.time() - start

        return loss, forward_time_first + forward_time_second, backward_time

    @staticmethod
    def _synchronize(device):
        if device.type == 'cuda':
            torch.cuda.synchronize()



import collections
import dataclasses
import logging
import random
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Few-shot sampling of a IC/OD dataset.
    For n-shot, do random sampling without replacement until each category exists in at least n images or all images are sampled.
    Given an image, add it if any classes shown in that image has less than n images sampled. e.g., in a dataset with 3 classes,
    for 1-shot, if already sampled 2 images and the # images by class is [0,1 2], only an image with class 0 object will be added.

    If config.strict is True and some classes have less than n_shot images, raise an error; otherwise sample all images from those classes.
    """
    VERSION = '0.2.1'

    @dataclasses.dataclass
    class Config:
        n_shot: int
        random_seed: int = 0
        strict: bool = False

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        few_shot_dataset = FewShotDataset(inputs.dataset, self.config.n_shot, self.config.random_seed, self.config.strict)
        return self.Outputs(few_shot_dataset)

    def dry_run(self, inputs):
        return self.execute(inputs)


def _get_class_id_set(targets):
    if not targets.shape:
        return {int(targets)}
    elif len(targets.shape) == 1:
        return set(int(t) for t in targets)
    elif len(targets.shape) == 2:
        return set(int(t[0]) for t in targets)
    raise ValueError(f"Unsupported target type is detected: {targets}")


class FewShotDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, n_shots, random_seed, strict=False):
        self._dataset = dataset

        classes_by_images = []
        for i in range(len(dataset)):
            target = dataset.get_targets(i) if hasattr(dataset, 'get_targets') else dataset[i][1]
            classes_by_images.append(_get_class_id_set(target))
        classes = set.union(*classes_by_images)
        classes_counter = collections.Counter({c: 0 for c in classes})
        num_images_by_class = collections.Counter([c for cs in classes_by_images for c in cs])
        if not strict:
            for c, n in num_images_by_class.items():
                if n < n_shots:
                    logger.debug(f'Class {c} only has {n} images, sample all {n}.')
        else:
            min_images_by_class = min(num_images_by_class.values())
            if not (0 < n_shots <= min_images_by_class):
                raise ValueError(f"n_shots must be in (0, {min_images_by_class}]")

        # Random sample until satisfying classes_freq.
        ids = list(range(len(dataset)))
        random.Random(random_seed).shuffle(ids)

        self._id_mappings = []
        for i in ids:
            # Only sample if any class shown in that image does not have enough samples yet.
            sample_needed = any([classes_counter[c] < n_shots for c in classes_by_images[i]])
            if sample_needed:
                self._id_mappings.append(i)
                classes_counter.update(classes_by_images[i])

        logger.info(f"Sampled {n_shots}-shot dataset with seed {random_seed}: {len(dataset)} -> {len(self._id_mappings)} samples.")

    def __len__(self):
        return len(self._id_mappings)

    def __getitem__(self, index):
        new_id = self._id_mappings[index]
        assert isinstance(new_id, int)
        return self._dataset[new_id]



import dataclasses
import torch
import irisml.core


class DatasetWithTargets(torch.utils.data.Dataset):
    def __init__(self, dataset, targets):
        super().__init__()
        self._dataset = dataset
        self._targets = targets

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, index):
        return self._dataset[index][0], self._targets[index]


class Task(irisml.core.TaskBase):
    """Make a classification dataset from predictions.

    This task takes a dataset and a tensor of predictions and returns a new dataset
    with the targets set to the predicted class.

    The predictions tensor can be either a 1D tensor of class indices or a 2D tensor
    of class probabilities. In the latter case, the class with the highest probability
    is used as the target.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset
        predictions: torch.Tensor  # Shape: (N, C) or (N,)

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        if len(inputs.predictions.shape) == 1:
            targets = inputs.predictions
        elif len(inputs.predictions.shape) == 2:
            targets = torch.argmax(inputs.predictions, dim=1)
        else:
            raise ValueError(f"Invalid shape for predictions: {inputs.predictions.shape}")

        if len(inputs.dataset) != len(targets):
            raise ValueError(f"Dataset and predictions have different lengths: {len(inputs.dataset)} vs {len(targets)}")

        return self.Outputs(dataset=DatasetWithTargets(inputs.dataset, targets))

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import logging
import typing
import torch

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class LrSchedulerConfig:
    name: typing.Literal['cosine_annealing', 'linear_decreasing', 'linear', 'multistep', 'multistep_with_warm_up'] = 'cosine_annealing'
    warmup_epochs: typing.Optional[int] = None
    warmup_iters: typing.Optional[int] = None
    warmup_lr: typing.Optional[float] = None
    warmup_factor: typing.Optional[float] = None
    warmup_end_factor: typing.Optional[float] = 1.0
    min_lr: typing.Optional[float] = 0.0  # For cosine_annealing LR scheduler
    end_factor: typing.Optional[float] = 1.0  # For linear LR scheduler
    decay_milestones: typing.Optional[typing.List[float]] = None  # For multistep LR scheduler
    decay_rate: typing.Optional[float] = 0.1  # For multistep LR scheduler


class LinearLR(torch.optim.lr_scheduler._LRScheduler):
    """Custom LinearLR implmentation. The official pytorch LinearLR doesn't support end_factor > 1."""
    def __init__(self, optimizer, end_factor, total_iters, last_epoch=-1, verbose=False):
        self._end_factor = end_factor
        self._total_iters = total_iters
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self):
        if self.last_epoch == 0 or self.last_epoch > self._total_iters:
            factor = 1
        else:
            factor = 1 + (self._end_factor - 1) * min(self._total_iters, self.last_epoch) / self._total_iters

        return [base_lr * factor for base_lr in self.base_lrs]


class MultiStepLRWithWarmUp(torch.optim.lr_scheduler.MultiStepLR):
    """Custom MultiStepLR implementation. The official pytorch MultiStepLR doesn't support warmup."""
    def __init__(self, optimizer, milestones, gamma, warmup_iters=0, warmup_lr_init=0, last_epoch=-1, verbose=False):
        self._warmup_iters = warmup_iters
        self._warmup_lr_init = warmup_lr_init
        assert self._warmup_iters > 0
        self._warmup_steps = [(v - warmup_lr_init) / self._warmup_iters for v in [group['lr'] for group in optimizer.param_groups]]
        super().__init__(optimizer, milestones, gamma, last_epoch, verbose)
        for param_group in optimizer.param_groups:
            param_group['lr'] = warmup_lr_init

    def get_lr(self):
        if self.last_epoch < self._warmup_iters:
            return [self._warmup_lr_init + self.last_epoch * s for s in self._warmup_steps]
        else:
            return super().get_lr()

    def _get_closed_form_lr(self):
        raise NotImplementedError


def build_lr_scheduler(config, optimizer, num_epochs, num_iterations_per_epoch):
    config = config or LrSchedulerConfig()
    total_iters = num_iterations_per_epoch * num_epochs
    warmup_iters = None

    if config.name == 'multistep_with_warm_up':
        return MultiStepLRWithWarmUp(optimizer,
                                     milestones=[int(total_iters*milestone) for milestone in config.decay_milestones],
                                     gamma=config.decay_rate,
                                     warmup_iters=num_iterations_per_epoch*config.warmup_epochs,
                                     warmup_lr_init=config.warmup_lr)

    if (config.warmup_epochs or config.warmup_iters) and config.warmup_factor:
        if config.warmup_epochs and config.warmup_iters:
            raise ValueError("warmup_epochs and warmup_iters cannot be used at the same time.")

        warmup_iters = num_iterations_per_epoch * config.warmup_epochs if config.warmup_epochs else config.warmup_iters
        logger.debug(f"Use learning rate warmup for {warmup_iters} iterations. {config.warmup_factor=}, {config.warmup_end_factor=}")
        # Make sure total_iters is a positive integer. Some LR scheduler doesn't allow negative values.
        total_iters = max(1, total_iters - warmup_iters)

    scheduler = None
    if config.name == 'cosine_annealing':
        scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, total_iters, eta_min=config.min_lr)
    elif config.name == 'linear_decreasing':
        scheduler = torch.optim.lr_scheduler.LinearLR(optimizer, start_factor=1, end_factor=0, total_iters=total_iters)
    elif config.name == 'linear':
        scheduler = LinearLR(optimizer, end_factor=config.end_factor, total_iters=total_iters)
    elif config.name == 'multistep':
        scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[int(total_iters*milestone) for milestone in config.decay_milestones], gamma=config.decay_rate)

    if not scheduler:
        raise ValueError(f"Unsupported lr scheduler name: {config.name}")

    if warmup_iters:
        warmup_scheduler = torch.optim.lr_scheduler.LinearLR(optimizer, start_factor=config.warmup_factor, end_factor=config.warmup_end_factor, total_iters=warmup_iters)
        scheduler = torch.optim.lr_scheduler.SequentialLR(optimizer, [warmup_scheduler, scheduler], milestones=[warmup_iters])

    return scheduler


class LrSchedulerFactory:
    def __init__(self, config: LrSchedulerConfig, num_epochs: int, num_iterations_per_epoch: int):
        self._config = config
        self._num_epochs = num_epochs
        self._num_iterations_per_epoch = num_iterations_per_epoch

    def __call__(self, optimizer):
        return build_lr_scheduler(self._config, optimizer, self._num_epochs, self._num_iterations_per_epoch)



import unittest
import torch
from irisml.tasks.get_subclass_dataset import Task


class TestSubClasstDataset(unittest.TestCase):
    def test_ic_multiclass(self):
        num_classes = 10
        dataset = [(None, torch.tensor(i % num_classes)) for i in range(50)]
        class_names = [str(i) for i in range(num_classes)]

        for class_id in range(num_classes):
            outputs = Task(Task.Config([class_id])).execute(Task.Inputs(dataset, class_names))
            self.assertEqual(len(outputs.dataset), 5)
            self.assertEqual(outputs.num_classes, 1)
            self.assertCountEqual(outputs.class_names, [str(class_id)])
            classes = set(int(t) for _, t in outputs.dataset)
            self.assertSetEqual(classes, {class_id})

        outputs = Task(Task.Config([0, 1])).execute(Task.Inputs(dataset, class_names))
        self.assertEqual(outputs.class_names, ['0', '1'])
        self.assertEqual(len(outputs.dataset), 10)
        self.assertEqual(outputs.num_classes, 2)

        with self.assertRaises(ValueError):
            outputs = Task(Task.Config([0, 12])).execute(Task.Inputs(dataset, class_names))

        with self.assertRaises(ValueError):
            outputs = Task(Task.Config([10])).execute(Task.Inputs(dataset, class_names))

    def test_od(self):
        dataset = [(None, torch.tensor([[0, 0, 0, 0.5, 0.5], [1, 0.5, 0.5, 1, 1]])), (None, torch.tensor([[1, 0, 0, 0.2, 0.2], [1, 0.2, 0.2, 0.4, 0.4]])), (None, torch.tensor([[0, 0, 0, 1, 1]]))]
        class_names = ['0', '1']
        outputs = Task(Task.Config([0])).execute(Task.Inputs(dataset, class_names))
        self.assertEqual(len(outputs.dataset), 2)
        self.assertCountEqual(outputs.class_names, ['0'])
        self.assertEqual(outputs.num_classes, 1)

        outputs = Task(Task.Config([1])).execute(Task.Inputs(dataset, class_names))
        self.assertEqual(len(outputs.dataset), 2)
        self.assertCountEqual(outputs.class_names, ['1'])

        with self.assertRaises(ValueError):
            outputs = Task(Task.Config([3, 0])).execute(Task.Inputs(dataset, class_names))



import dataclasses
import typing
import irisml.core


def clip_imagenet_generator(label_name: str) -> typing.List[str]:
    templates = [
        'a bad photo of a {}.',
        'a photo of many {}.',
        'a sculpture of a {}.',
        'a photo of the hard to see {}.',
        'a low resolution photo of the {}.',
        'a rendering of a {}.',
        'graffiti of a {}.',
        'a bad photo of the {}.',
        'a cropped photo of the {}.',
        'a tattoo of a {}.',
        'the embroidered {}.',
        'a photo of a hard to see {}.',
        'a bright photo of a {}.',
        'a photo of a clean {}.',
        'a photo of a dirty {}.',
        'a dark photo of the {}.',
        'a drawing of a {}.',
        'a photo of my {}.',
        'the plastic {}.',
        'a photo of the cool {}.',
        'a close-up photo of a {}.',
        'a black and white photo of the {}.',
        'a painting of the {}.',
        'a painting of a {}.',
        'a pixelated photo of the {}.',
        'a sculpture of the {}.',
        'a bright photo of the {}.',
        'a cropped photo of a {}.',
        'a plastic {}.',
        'a photo of the dirty {}.',
        'a jpeg corrupted photo of a {}.',
        'a blurry photo of the {}.',
        'a photo of the {}.',
        'a good photo of the {}.',
        'a rendering of the {}.',
        'a {} in a video game.',
        'a photo of one {}.',
        'a doodle of a {}.',
        'a close-up photo of the {}.',
        'a photo of a {}.',
        'the origami {}.',
        'the {} in a video game.',
        'a sketch of a {}.',
        'a doodle of the {}.',
        'a origami {}.',
        'a low resolution photo of a {}.',
        'the toy {}.',
        'a rendition of the {}.',
        'a photo of the clean {}.',
        'a photo of a large {}.',
        'a rendition of a {}.',
        'a photo of a nice {}.',
        'a photo of a weird {}.',
        'a blurry photo of a {}.',
        'a cartoon {}.',
        'art of a {}.',
        'a sketch of the {}.',
        'a embroidered {}.',
        'a pixelated photo of a {}.',
        'itap of the {}.',
        'a jpeg corrupted photo of the {}.',
        'a good photo of a {}.',
        'a plushie {}.',
        'a photo of the nice {}.',
        'a photo of the small {}.',
        'a photo of the weird {}.',
        'the cartoon {}.',
        'art of the {}.',
        'a drawing of the {}.',
        'a photo of the large {}.',
        'a black and white photo of a {}.',
        'the plushie {}.',
        'a dark photo of a {}.',
        'itap of a {}.',
        'graffiti of the {}.',
        'a toy {}.',
        'itap of my {}.',
        'a photo of a cool {}.',
        'a photo of a small {}.',
        'a tattoo of the {}.'
        ]
    return [t.format(label_name) for t in templates]


def clip_imagenet_short_generator(label_name: str) -> typing.List[str]:
    templates = ['itap of a {}.',
                 'a bad photo of the {}.',
                 'a origami {}.',
                 'a photo of the large {}.',
                 'a {} in a video game.',
                 'art of the {}.',
                 'a photo of the small {}.']
    return [t.format(label_name) for t in templates]


class Task(irisml.core.TaskBase):
    """Create a prompt generator for a classification task.

    Supported generator:
        'clip_imagenet': Generates 80 prompts that are introduced in the openai/CLIP repository. (https://github.com/openai/CLIP)
        'clip_imagenet_short': Generates 7 prompts that are introduced in the openai/CLIP repository.
    """
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Config:
        name: str = 'clip_imagenet'

    @dataclasses.dataclass
    class Outputs:
        generator: typing.Callable[[str], typing.List[str]]

    GENERATORS = {
        'clip_imagenet': clip_imagenet_generator,
        'clip_imagenet_short': clip_imagenet_short_generator
    }

    def execute(self, inputs):
        if self.config.name not in self.GENERATORS:
            raise ValueError(f"Unknown generator name: {self.config.name}")
        gen = self.GENERATORS[self.config.name]
        return self.Outputs(gen)

    def dry_run(self, inputs):
        return self.execute(inputs)



import copy
import dataclasses
import logging
import random
import time
import typing
import torch
import irisml.core

from .build_dataloader import build_dataloader
from .build_lr_scheduler import LrSchedulerFactory, LrSchedulerConfig
from .build_optimizer import OptimizerFactory
from .ddp_trainer import DDPTrainer
from .plugin_loader import load_plugin
from .trainer import Trainer

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Train a pytorch model.

    This is a simple training baseline.

    Config:
        num_epochs (int): The number of epochs.
        batch_size (int): Batch size.
        base_lr (float): Base learning rate.
        device ('cpu' or 'cuda'): Device for training.
        lr_scheduler (LrSchedulerConfig): Learning rate scheduler config.
        momentum (float): Momentum parameter for optimizer.
        val_check_interval(float or int): Validation phase interval.
        weight_decay (float): Weight decay.
        num_processes (int): The number of processes to use. Default: 1.
        no_weight_decay_param_names (Optional[List[str]]): A regex pattern for parameter names to disable weight_decay
        no_weight_decay_module_class_names (Optional[List[str]]): A regex pattern for module class names to disable weight_decay
        lr_scale_param_names (Optional[List[Tuple[str, float]]]): Learning rate scaling for matched parameters.
        plugins (List[str]): Plugins to use for training. Each plugin can have constant values as arguments. e.g. "clip_max_norm(10)".
        random_seed (Optional[int]): Reset the random seed if requested
    """
    VERSION = '0.2.9'

    @dataclasses.dataclass
    class Config:
        num_epochs: int
        batch_size: int = 1
        base_lr: float = 1e-5
        device: typing.Optional[typing.Literal['cpu', 'cuda']] = None
        lr_scheduler: typing.Optional[LrSchedulerConfig] = None
        accumulate_grad_batches: typing.Optional[int] = None
        momentum: float = 0
        optimizer: typing.Literal['sgd', 'adam', 'amsgrad', 'adamw', 'adamw_amsgrad', 'rmsprop', 'zero_adam', 'zero_amsgrad', 'zero_adamw', 'zero_adamw_amsgrad'] = 'sgd'
        val_check_interval: typing.Optional[typing.Union[int, float]] = None
        weight_decay: float = 0
        num_processes: int = 1
        no_weight_decay_param_names: typing.Optional[typing.List[str]] = None
        no_weight_decay_module_class_names: typing.Optional[typing.List[str]] = None
        lr_scale_param_names: typing.Optional[typing.List[typing.Tuple[str, float]]] = None
        plugins: typing.List[str] = dataclasses.field(default_factory=list)
        random_seed: typing.Optional[int] = None

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module
        train_dataset: torch.utils.data.Dataset
        train_transform: typing.Callable
        val_dataset: torch.utils.data.Dataset = None
        val_transform: typing.Callable = None

    @dataclasses.dataclass
    class Outputs:
        model: torch.nn.Module = None
        elapsed_time: float = 0

    def execute(self, inputs):
        start_time = time.time()
        model = self.train(inputs)
        return self.Outputs(model, time.time() - start_time)

    def dry_run(self, inputs):
        start_time = time.time()
        model = self.train(inputs, dry_run=True)
        return self.Outputs(model, time.time() - start_time)

    def train(self, inputs, dry_run=False) -> torch.nn.Module:
        if self.config.random_seed:  # Reset the random seed if requested. Note that the IRISML platform also resets the seed at the beginning of each task.
            torch.manual_seed(self.config.random_seed)
            random.seed(self.config.random_seed)
            logger.info(f"Reset the random seed to {self.config.random_seed}")

        plugins = [load_plugin(p, self.context) for p in self.config.plugins]
        if dry_run:
            train_dataloader = val_dataloader = None
        else:
            train_dataloader = build_dataloader(inputs.train_dataset, inputs.train_transform, batch_size=self.config.batch_size, num_processes=self.config.num_processes)
            val_dataloader = inputs.val_dataset and build_dataloader(inputs.val_dataset, inputs.val_transform, batch_size=self.config.batch_size, shuffle=False, drop_last=False)
        device = self._get_device()
        model = copy.deepcopy(inputs.model) if not dry_run else inputs.model  # Save CPU memory on dry run

        optimizer_factory = OptimizerFactory(self.config.optimizer, self.config.base_lr, self.config.weight_decay, self.config.momentum,
                                             self.config.no_weight_decay_param_names, self.config.no_weight_decay_module_class_names,
                                             self.config.lr_scale_param_names)
        lr_scheduler_factory = LrSchedulerFactory(self.config.lr_scheduler, self.config.num_epochs, len(train_dataloader) if train_dataloader else 1)

        trainer_parameters = {'model': model, 'lr_scheduler_factory': lr_scheduler_factory, 'optimizer_factory': optimizer_factory, 'plugins': plugins,
                              'device': device, 'val_check_interval': self.config.val_check_interval, 'accumulate_grad_batches': self.config.accumulate_grad_batches}

        trainer = self.get_trainer(trainer_parameters)

        if not dry_run:
            trainer.train(train_dataloader, val_dataloader, num_epochs=self.config.num_epochs)

        return trainer.model

    def get_trainer(self, parameters):
        """Get a trainer instance. This method may be overridden by a subclass."""
        if self.config.num_processes == 1:
            return Trainer(**parameters)
        else:
            return DDPTrainer(num_processes=self.config.num_processes, **parameters)

    def _get_device(self) -> torch.device:
        """Get a torch device based on the configuration. If not specified explicitly, it uses cuda if available."""
        if self.config.device:
            device_name = self.config.device
        else:
            device_name = 'cuda' if torch.cuda.is_available() else 'cpu'
            logger.info(f"Training device is selected automatically: {device_name}. To specify the device manually, please set Config.device.")

        return torch.device(device_name)



import collections
import unittest
import torch
from irisml.tasks.sample_few_shot_dataset import Task

from utils import FakeDatasetWithGetTargets


class TestSampleFewShotDataset(unittest.TestCase):
    def test_ic_multiclass_few_shot(self):
        dataset = [(None, torch.tensor(i % 10)) for i in range(100)]
        for i in range(10):
            outputs = Task(Task.Config(1, random_seed=i)).execute(Task.Inputs(dataset))
            self.assertGreaterEqual(len(outputs.dataset), 10)
            classes = set(int(t) for _, t in outputs.dataset)
            self.assertEqual(len(classes), 10)

    def test_ic_multiclass_few_shot_rare_classes(self):
        """Test some classes have less than n_shot images."""
        dataset = [(None, torch.tensor(c)) for c in [0, 1, 1, 1, 2, 2, 2, 2]]
        for i in range(10):
            outputs = Task(Task.Config(n_shot=2, random_seed=i, strict=False)).execute(Task.Inputs(dataset))
            # Can only sample 1 image for class 0, and 2 for class 1,2
            self.assertEqual(len(outputs.dataset), 5)
            classes = set(int(t) for _, t in outputs.dataset)
            self.assertEqual(len(classes), 3)

    def test_default_seed(self):
        dataset = [(None, torch.tensor(i % 10)) for i in range(100)]
        all_classes = []
        for i in range(10):
            outputs = Task(Task.Config(1)).execute(Task.Inputs(dataset))
            classes = [int(t) for _, t in outputs.dataset]
            all_classes.append(classes)
            if i > 0:
                self.assertEqual(classes, all_classes[0])

    def test_od_few_shot(self):
        dataset = [(None, torch.tensor([[0, 0, 0, 0.5, 0.5], [1, 0.5, 0.5, 1, 1]])), (None, torch.tensor([[1, 0, 0, 0.2, 0.2], [1, 0.2, 0.2, 0.4, 0.4]])), (None, torch.tensor([[0, 0, 0, 1, 1]]))]
        outputs = Task(Task.Config(2)).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs.dataset), 3)

        classes_counter = collections.Counter()
        for _, targets in dataset:
            classes_counter.update(set(int(t[0]) for t in targets))

        for n_images in classes_counter.values():
            self.assertGreaterEqual(n_images, 2)

        dataset = [(None, torch.tensor([[0, 0, 0, 0.5, 0.5], [1, 0.5, 0.5, 1, 1]])),
                   (None, torch.tensor([[0, 0, 0, 0.2, 0.2]])), (None, torch.tensor([[0, 0, 0, 1, 1]])), (None, torch.tensor([[0, 0, 0, 1, 1]]))]
        outputs = Task(Task.Config(1)).execute(Task.Inputs(FakeDatasetWithGetTargets(dataset)))
        self.assertLessEqual(len(outputs.dataset), 2)

    def test_ic_multiclass_get_targets(self):
        dataset = FakeDatasetWithGetTargets([(None, torch.tensor([1])), (None, torch.tensor([1])), (None, torch.tensor([0]))])
        outputs = Task(Task.Config(1)).execute(Task.Inputs(dataset))
        self.assertGreaterEqual(len(outputs.dataset), 2)



import typing
import dataclasses
import logging
import torch.utils.data

import irisml.core
from irisml.tasks.sample_few_shot_dataset import _get_class_id_set

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Get the sub-dataset with given class ids from a dataset."""
    VERSION = '0.0.1'

    @dataclasses.dataclass
    class Config:
        class_ids: typing.List[int]

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset
        class_names: typing.List[str]

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset
        class_names: typing.List[str]
        num_classes: int

    def execute(self, inputs):
        class_ids = sorted(self.config.class_ids)
        subclass_dataset = SubClassDataset(inputs.dataset, class_ids)
        subclass_names = [inputs.class_names[i] for i in class_ids]
        return Task.Outputs(subclass_dataset, subclass_names, len(subclass_names))

    def dry_run(self, inputs):
        return self.execute(inputs)


class SubClassDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, class_ids):
        super().__init__()
        self._dataset = dataset
        subclasses = set(class_ids)

        classes_by_images = []
        for i in range(len(dataset)):
            target = dataset.get_targets(i) if hasattr(dataset, 'get_targets') else dataset[i][1]
            classes_by_images.append(_get_class_id_set(target))
        classes = set.union(*classes_by_images)

        if not subclasses.issubset(classes):
            raise ValueError(f"Given sub-class ids {subclasses} are not a subset of dataset class ids: {classes}.")
        self._id_mappings = [i for i in range(len(dataset)) if classes_by_images[i].intersection(subclasses)]
        logger.info(f"Extracted subset with {len(self._id_mappings)} images of given class ids: {class_ids}.")

    def __len__(self):
        return len(self._id_mappings)

    def __getitem__(self, index):
        new_id = self._id_mappings[index]
        assert isinstance(new_id, int)
        return self._dataset[new_id]

    def get_targets(self, index):
        return self._dataset.get_targets(self._id_mappings[index])



import dataclasses
import itertools
import logging
import math
import multiprocessing
import os
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Extract only targets from a given Dataset.

    If the targets are integers or tensors with same shape, a tensor will be returned. Otherwise, returns a list of targets.

    If the dataset instance provides 'get_targets(index)' method, this task will use it in order to get targets faster.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        targets: typing.Union[typing.List, torch.Tensor] = dataclasses.field(default_factory=list)

    def execute(self, inputs):
        targets = self._collect_targets(inputs.dataset)
        if torch.is_tensor(targets[0]):
            shape = targets[0].shape
            if all(t.shape == shape for t in targets):
                targets = torch.stack(targets)
        elif isinstance(targets[0], int):
            targets = torch.tensor(targets)

        return self.Outputs(targets)

    @staticmethod
    def _worker(dataset, process_index, batch_size, queue):
        start_index = batch_size * process_index
        end_index = min(start_index + batch_size, len(dataset))
        targets = [dataset[i][1] for i in range(start_index, end_index)]
        queue.put((process_index, targets))
        queue.join()

    @staticmethod
    def _collect_targets(dataset):
        if hasattr(dataset, 'get_targets'):
            return [dataset.get_targets(i) for i in range(len(dataset))]

        try:
            num_processes = len(os.sched_getaffinity(0))
            assert num_processes > 0
            batch_size = math.ceil(len(dataset) / num_processes)

            logger.info(f"Using {num_processes} processes to collect {len(dataset)} targets.")

            context = multiprocessing.get_context('fork')
            queue = context.JoinableQueue(num_processes)
            processes = [context.Process(target=Task._worker, args=(dataset, i, batch_size, queue), daemon=True) for i in range(num_processes)]
            for p in processes:
                p.start()

            collected_batches = [None] * num_processes
            for _ in range(num_processes):
                i, batch = queue.get(timeout=3600)
                collected_batches[i] = batch
                queue.task_done()

            for p in processes:
                p.join()

            assert all(b is not None for b in collected_batches)

            targets = list(itertools.chain.from_iterable(collected_batches))
            assert len(targets) == len(dataset)
            return targets

        except Exception:
            logger.exception("Failed to collect targets in parallel.")
            return [t for _, t in dataset]



import copy
import dataclasses
import logging
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Concatenate the given two datasets together.

    If Config.use_same_classspace is False, class_names0 and class_names1 must be provided. The class ids will be mapped so that the new class_names will not have duplicated names.

    Config:
        task_type ('classifictaion_multiclass' or 'object_detection'): The dataset type.
        use_same_classspace (bool): If False, the class ids of the dataset1 will be mapped to new class ids.
    """
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Inputs:
        dataset0: torch.utils.data.Dataset
        dataset1: typing.Optional[torch.utils.data.Dataset] = None
        class_names0: typing.Optional[typing.List[str]] = None
        class_names1: typing.Optional[typing.List[str]] = None

    @dataclasses.dataclass
    class Config:
        task_type: typing.Literal['classification_multiclass', 'object_detection'] = 'classification_multiclass'
        use_same_classspace: bool = True

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset
        class_names: typing.Optional[typing.List[str]] = None
        num_classes: typing.Optional[int] = None

    def execute(self, inputs):
        if not inputs.dataset1:
            logger.info("dataset1 is None. Returning dataset0 without concatenation.")
            return self.Outputs(inputs.dataset0, inputs.class_names0)

        if not self.config.use_same_classspace:
            if inputs.class_names0 and inputs.class_names1:
                class_mappings = [list(range(len(inputs.class_names0))), []]
                new_class_names = copy.deepcopy(inputs.class_names0)
                for c in inputs.class_names1:
                    try:
                        existing_class_id = new_class_names.index(c)
                        class_mappings[1].append(existing_class_id)
                    except ValueError:
                        class_mappings[1].append(len(new_class_names))
                        new_class_names.append(c)
            else:
                raise RuntimeError("When use_same_classspace is False, both class_names0 and class_names1 must be provided.")
            num_classes = len(new_class_names)
            logger.info(f"The number of new classes is {num_classes}")
        else:
            class_mappings = None
            new_class_names = None
            num_classes = None

        dataset = ConcatDataset(inputs.dataset0, inputs.dataset1, class_mappings, self.config.task_type)
        return self.Outputs(dataset, new_class_names, num_classes)

    def dry_run(self, inputs):
        return self.execute(inputs)


class ConcatDataset(torch.utils.data.Dataset):
    def __init__(self, dataset0, dataset1, class_mappings: typing.List[typing.List[int]], task_type):
        assert class_mappings is None or len(class_mappings) == 2
        assert task_type in ['classification_multiclass', 'object_detection']
        self._datasets = [dataset0, dataset1]
        self._first_dataset_size = len(dataset0)
        self._class_mappings = class_mappings
        self._task_type = task_type

    def __len__(self):
        return len(self._datasets[0]) + len(self._datasets[1])

    def __getitem__(self, index):
        if index < self._first_dataset_size:
            dataset_index = 0
            sample_index = index
        else:
            dataset_index = 1
            sample_index = index - self._first_dataset_size

        inputs, targets = self._datasets[dataset_index][sample_index]

        if self._class_mappings:
            if self._task_type == 'classification_multiclass':
                if isinstance(targets, list) and len(targets) == 1:
                    targets = targets[0]
                targets = torch.tensor([self._class_mappings[dataset_index][int(targets)]])
            elif self._task_type == 'object_detection':
                targets = torch.tensor([[self._class_mappings[dataset_index][int(t[0])], *t[1:]] for t in targets])

        return inputs, targets



import dataclasses
import logging
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """
    Remove empty images from dataset.
    Currently this task is only needed by FL OD Dyhead model training
    """
    VERSION = '0.0.1'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        original_dataset_size = len(inputs.dataset)

        dataset = EmptyRemovedDataset(inputs.dataset)

        logger.info(f"Removing images without any annotations."
                    f"The number of left samples is {len(dataset)}. The original number of samples is {original_dataset_size}.")

        return self.Outputs(dataset)

    def dry_run(self, inputs):
        return self.Outputs(inputs.dataset)


class EmptyRemovedDataset(torch.utils.data.Dataset):
    def __init__(self, dataset):
        self._dataset = dataset

        self._id_mappings = []
        for i in range(len(dataset)):
            if hasattr(dataset, 'get_targets'):
                targets = dataset.get_targets(i)
            else:
                targets = dataset[i][1]
            if not targets.shape or len(targets):
                self._id_mappings.append(i)

    def __len__(self):
        return len(self._id_mappings)

    def __getitem__(self, index):
        new_id = self._id_mappings[index]
        assert isinstance(new_id, int)
        return self._dataset[new_id]



import dataclasses
import logging
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Find incorrect classification indices."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        predictions: torch.Tensor
        targets: torch.Tensor

    @dataclasses.dataclass
    class Outputs:
        indices: torch.Tensor

    def execute(self, inputs):
        if len(inputs.predictions) != len(inputs.targets):
            raise ValueError("The number of predictions and targets must be the same.")

        if len(inputs.predictions.shape) == 1:
            predicted_max_indices = inputs.predictions
        elif len(inputs.predictions.shape) == 2:
            predicted_max_indices = torch.argmax(inputs.predictions, dim=1)
        else:
            raise ValueError("The predictions tensor must be 1D or 2D.")

        assert predicted_max_indices.shape == inputs.targets.shape

        indices = []
        for i in range(len(predicted_max_indices)):
            if predicted_max_indices[i] != inputs.targets[i]:
                logger.info(f"Index {i}: Predicted: {predicted_max_indices[i]}, Actual: {inputs.targets[i]}")

                indices.append(i)
        return self.Outputs(torch.tensor(indices))

    def dry_run(self, inputs):
        return self.execute(inputs)



"""Use PyTorch Autograd's anomaly detection.

Useful for debugging a training issue. Note that this plugin is not recommended for production. It will slow down your training.
"""
import logging
import torch.autograd
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self):
        logger.info("Anomaly Detection is enabled. It will slow down the training.")

    def on_train_start(self, trainer, model):
        torch.autograd.set_detect_anomaly(True)

    def on_train_end(self, trainer, model):
        torch.autograd.set_detect_anomaly(False)



import dataclasses
import typing
import PIL.Image
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Make a transform function for image and a fixed prompt.

    The interface of the transform function is:
    - Input: PIL.Image.Image
    - Output: Tuple[str, List[torch.Tensor]]

    Config:
        prompt (str): The prompt to use.
    """

    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        image_transform: typing.Callable[[PIL.Image.Image], torch.Tensor]

    @dataclasses.dataclass
    class Config:
        prompt: str

    @dataclasses.dataclass
    class Outputs:
        transform: typing.Callable

    def execute(self, inputs):
        transform = FixedPromptImageTransform(inputs.image_transform, self.config.prompt)
        return self.Outputs(transform=transform)

    def dry_run(self, inputs):
        return self.execute(inputs)


class FixedPromptImageTransform:
    def __init__(self, image_transform, prompt):
        self._image_transform = image_transform
        self._prompt = prompt

    def __call__(self, inputs, targets):
        assert isinstance(inputs, PIL.Image.Image)
        image_tensor = self._image_transform(inputs)
        assert isinstance(image_tensor, torch.Tensor)
        return (self._prompt, [image_tensor]), targets



import dataclasses
import logging
import torch
import irisml.core

logger = logging.getLogger(__name__)


class RemappedDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, mappings):
        self._dataset = dataset
        self._mappings = mappings
        if hasattr(dataset, 'get_targets'):
            self.get_targets = self.__get_targets

    def __len__(self):
        return len(self._mappings)

    def __getitem__(self, index):
        return self._dataset[self._mappings[index]]

    def __get_targets(self, index):
        return self._dataset.get_targets(self._mappings[index])


class Task(irisml.core.TaskBase):
    """Exclude negative samples from classification dataset."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        mappings = self._get_mappings(inputs.dataset)
        logger.info(f"New total number of samples: {len(mappings)}. Excluded {len(inputs.dataset) - len(mappings)} samples.")
        return self.Outputs(dataset=RemappedDataset(inputs.dataset, mappings))

    def dry_run(self, inputs):
        return self.execute(inputs)

    @staticmethod
    def _get_mappings(dataset):
        mappings = []
        for i in range(len(dataset)):
            if hasattr(dataset, 'get_targets'):
                targets = dataset.get_targets(i)
            else:
                targets = dataset[i][1]
            if int(targets) < 0:
                continue
            mappings.append(i)
        return mappings



import unittest
import torch
from irisml.tasks.make_oversampled_dataset import Task

from utils import FakeDatasetWithGetTargets


class TestMakeOversampledDataset(unittest.TestCase):
    def test_calculate_weights(self):
        dataset = [(None, torch.tensor([1])), (None, torch.tensor([1])), (None, torch.tensor([0]))]
        results = Task._calculcate_weights(FakeDatasetWithGetTargets(dataset), 100, 0)
        self.assertEqual(results[0], results[1])
        self.assertGreater(results[2], results[0])

    def test_min_num_samples(self):
        dataset = [(None, torch.tensor([1])), (None, torch.tensor([1])), (None, torch.tensor([0]))]
        outputs = Task(Task.Config(10)).execute(Task.Inputs(FakeDatasetWithGetTargets(dataset)))
        self.assertEqual(len(outputs.dataset), 10)
        self.assertIsNotNone(outputs.dataset[0])
        self.assertIsNotNone(outputs.dataset[9])

        dataset = [(None, torch.tensor([1])), (None, torch.tensor([1])), (None, torch.tensor([0]))]
        dataset = FakeDatasetWithGetTargets(dataset)
        outputs = Task(Task.Config(3)).execute(Task.Inputs(dataset))
        self.assertEqual(outputs.dataset, dataset)  # no oversampling

    def test_oversampling_rate(self):
        dataset = [(None, torch.tensor([1])), (None, torch.tensor([1])), (None, torch.tensor([0]))]
        outputs = Task(Task.Config(0, 3)).execute(Task.Inputs(FakeDatasetWithGetTargets(dataset)))
        self.assertEqual(len(outputs.dataset), 9)

        dataset = [(None, torch.tensor([1])), (None, torch.tensor([1])), (None, torch.tensor([0]))]
        dataset = FakeDatasetWithGetTargets(dataset)
        outputs = Task(Task.Config(0, 1)).execute(Task.Inputs(dataset))
        self.assertEqual(outputs.dataset, dataset)  # no oversampling

    def test_balancing(self):
        dataset = [(None, torch.tensor([1])), (None, torch.tensor([1])), (None, torch.tensor([0]))]
        outputs = Task(Task.Config(0, 3, balance=True)).execute(Task.Inputs(FakeDatasetWithGetTargets(dataset)))
        self.assertEqual(len(outputs.dataset), 9)



import dataclasses
import logging
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Create a zero-shot classification layer."""
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Inputs:
        text_features: typing.List[torch.Tensor]
        text_classes: typing.Union[typing.List[int], torch.Tensor]
        logit_scale: typing.Optional[torch.Tensor] = None

    @dataclasses.dataclass
    class Config:
        num_classes: int

    @dataclasses.dataclass
    class Outputs:
        classifier: torch.nn.Module = None

    def execute(self, inputs):
        num_examples = len(inputs.text_features)
        if len(inputs.text_classes) != num_examples:
            raise RuntimeError(f"The number of examples doesn't match. features={num_examples}, classes={len(inputs.text_classes)}")

        feature_shape = inputs.text_features[0].shape
        logger.debug(f"Feature shape is {feature_shape}. Got {num_examples} samples.")

        features_per_class = [[] for _ in range(self.config.num_classes)]
        for c, f in zip(inputs.text_classes, inputs.text_features):
            index = int(c)
            if not 0 <= index < self.config.num_classes:
                raise RuntimeError(f"Unexpected class index: {index}")
            features_per_class[index].append(f)

        embeddings_per_class = []
        for c in range(self.config.num_classes):
            if features_per_class[c]:
                class_embeddings = torch.stack(features_per_class[c])
                class_embedding = class_embeddings.mean(dim=0)
                class_embedding /= class_embedding.norm()
            else:
                logger.warning(f"No features are provided for class {c}. Initializing with zeros.")
                class_embedding = torch.zeros(*feature_shape)

            embeddings_per_class.append(class_embedding)

        weights = torch.stack(embeddings_per_class, dim=1).transpose(0, 1)
        if inputs.logit_scale:
            if inputs.logit_scale.nelement() != 1:
                raise RuntimeError("Unexpected shape of logit_scale")
            weights *= inputs.logit_scale.exp()

        with torch.no_grad():
            classifier = torch.nn.Linear(weights.shape[1], weights.shape[0], bias=False)
            classifier.weight.copy_(weights)
        return self.Outputs(classifier)



import unittest
import torch
from irisml.tasks.predict import Task, SubsetSequentialSampler
from utils import FakeDataset


class FakeModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.model = torch.nn.Conv2d(3, 3, 3)

    def forward(self, x):
        return torch.flatten(torch.nn.AdaptiveAvgPool2d(1)(self.model(x)), start_dim=1)

    @property
    def predictor(self):
        return torch.nn.Softmax(1)


class FakeModelWithResult(torch.nn.Module):
    def __init__(self, data):
        super().__init__()
        self._data = data
        self._counter = 0

    def forward(self, x):
        x = self._data[self._counter]
        self._counter += 1
        return x

    @property
    def predictor(self):
        return torch.nn.Identity()


def fake_transform(x, y):
    return x, y


class TestPredict(unittest.TestCase):
    def test_single(self):
        dataset = FakeDataset([(torch.rand(3, 256, 256), torch.tensor(1)), (torch.rand(3, 256, 256), torch.tensor(2))])
        model = FakeModel()

        self._assert_prediction(dataset, model, 1)
        self._assert_prediction(dataset, model, 2)
        self._assert_prediction(dataset, model, 3)

    def test_no_transform(self):
        dataset = FakeDataset([(torch.rand(3, 256, 256), torch.tensor(1)), (torch.rand(3, 256, 256), torch.tensor(2))])
        model = FakeModel()
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset=dataset, model=model))
        self.assertEqual(len(outputs.predictions), len(dataset))
        self.assertEqual(len(outputs.targets), len(dataset))

    def test_distributed(self):
        dataset = FakeDataset([(torch.rand(3, 8, 8), torch.tensor(1)) for _ in range(6)])
        model = FakeModel()

        self._assert_prediction(dataset, model, 2, 2)
        self._assert_prediction(dataset, model, 1, 3)
        self._assert_prediction(dataset, model, 5, 2)

    def test_error_in_distributed(self):
        dataset = FakeDataset([(torch.rand(3, 8, 8), torch.tensor(1)) for _ in range(6)])
        model = FakeModel()
        model.forward = None  # It will fail to run the model.

        with self.assertRaises(Exception):
            self._assert_prediction(dataset, model, 2, 2)

    def test_sampler(self):
        dataset = FakeDataset([(torch.rand(3, 8, 8), torch.tensor(1)) for _ in range(6)])
        sampler = SubsetSequentialSampler(dataset, 0, 2)
        self.assertEqual(len(sampler), 2)
        iterator = iter(sampler)
        self.assertEqual(next(iterator), 0)
        self.assertEqual(next(iterator), 1)
        with self.assertRaises(StopIteration):
            next(iterator)

    def test_aggregate(self):
        # Tensors with same shape.
        dataset = FakeDataset([(torch.rand(3, 256, 256), torch.tensor(1)), (torch.rand(3, 256, 256), torch.tensor(2))])
        model = FakeModelWithResult([torch.tensor([1]), torch.tensor([2])])
        inputs = Task.Inputs(dataset=dataset, transform=fake_transform, model=model)
        outputs = Task(Task.Config(batch_size=1, num_processes=1)).execute(inputs)
        self.assertIsInstance(outputs.predictions, torch.Tensor)
        self.assertIsInstance(outputs.targets, torch.Tensor)
        self.assertEqual(outputs.predictions.tolist(), [1, 2])

        # Tensors with different shape.
        model = FakeModelWithResult([torch.tensor([[1]]), torch.tensor([[2, 3]])])
        inputs = Task.Inputs(dataset=dataset, transform=fake_transform, model=model)
        outputs = Task(Task.Config(batch_size=1, num_processes=1)).execute(inputs)
        self.assertIsInstance(outputs.predictions, list)
        self.assertEqual(len(outputs.predictions), 2)
        self.assertTrue(torch.equal(outputs.predictions[0], torch.tensor([1])))
        self.assertTrue(torch.equal(outputs.predictions[1], torch.tensor([2, 3])))

        # Lists
        model = FakeModelWithResult([[[1]], [[[2, 3], [4, 5]]]])
        inputs = Task.Inputs(dataset=dataset, transform=fake_transform, model=model)
        outputs = Task(Task.Config(batch_size=1, num_processes=1)).execute(inputs)
        self.assertIsInstance(outputs.predictions, list)
        self.assertEqual(len(outputs.predictions), 2)
        self.assertEqual(outputs.predictions, [[1], [[2, 3], [4, 5]]])

    def test_callback(self):
        dataset = FakeDataset([(torch.rand(3, 256, 256), torch.tensor(1)), (torch.rand(3, 256, 256), torch.tensor(2))])
        model = FakeModel()

        inputs = Task.Inputs(dataset=dataset, transform=fake_transform, model=model)
        outputs = Task(Task.Config(plugins=['heartbeat'])).execute(inputs)
        self.assertEqual(len(outputs.predictions), len(dataset))

    def _assert_prediction(self, dataset, model, batch_size, num_processes=1):
        inputs = Task.Inputs(dataset=dataset, transform=fake_transform, model=model)
        task = Task(Task.Config(batch_size=batch_size, num_processes=num_processes, device='cpu'))
        outputs = task.execute(inputs)
        self.assertEqual(len(outputs.predictions), len(dataset))
        self.assertEqual(len(outputs.targets), len(dataset))



import unittest
import torch
from irisml.tasks.concatenate_datasets import Task


from utils import FakeDataset


class TestConcatenateDatasets(unittest.TestCase):
    def test_same_classsspace_classification(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        dataset1 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        outputs = Task(Task.Config(use_same_classspace=True)).execute(Task.Inputs(dataset0, dataset1))
        self.assertEqual(len(outputs.dataset), 6)
        self.assertEqual(outputs.dataset[0], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[3], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[5], (None, torch.tensor(3)))

    def test_different_classsspace_classification_with_same_name(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        dataset1 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        class_names0 = ['label0', 'label1', 'label2', 'label3', 'label4']
        class_names1 = ['label0', 'label1', 'label2', 'label3', 'label4']
        outputs = Task(Task.Config(use_same_classspace=False)).execute(Task.Inputs(dataset0, dataset1, class_names0, class_names1))
        self.assertEqual(len(outputs.dataset), 6)
        self.assertEqual(outputs.dataset[0], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[3], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[5], (None, torch.tensor(3)))

    def test_different_classsspace_classification(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        dataset1 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        class_names0 = ['label0', 'label1', 'label2', 'label3', 'label4']
        class_names1 = ['label00', 'label01', 'label02', 'label03', 'label04']
        outputs = Task(Task.Config(use_same_classspace=False)).execute(Task.Inputs(dataset0, dataset1, class_names0, class_names1))
        self.assertEqual(len(outputs.dataset), 6)
        self.assertEqual(outputs.dataset[0], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[3], (None, torch.tensor(6)))
        self.assertEqual(outputs.dataset[5], (None, torch.tensor(8)))

    def test_different_classsspace_classification_with_duplicated_names(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        dataset1 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        class_names0 = ['label0', 'label1', 'label2', 'label3', 'label4']
        class_names1 = ['label0', 'label1', 'label02', 'label03', 'label04']
        outputs = Task(Task.Config(use_same_classspace=False)).execute(Task.Inputs(dataset0, dataset1, class_names0, class_names1))
        self.assertEqual(len(outputs.dataset), 6)
        self.assertEqual(outputs.dataset[0], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[3], (None, torch.tensor(1)))
        self.assertEqual(outputs.dataset[5], (None, torch.tensor(6)))

    def test_different_classsspace_object_detection_with_duplicated_names(self):
        dataset0 = FakeDataset([(None, torch.tensor([[1, 0, 0, 1, 1]])), (None, torch.tensor([[2, 0, 0, 1, 1]])), (None, torch.tensor([[3, 0, 0, 1, 1]]))])
        dataset1 = FakeDataset([(None, torch.tensor([[1, 0, 0, 1, 1]])), (None, torch.tensor([[2, 0, 0, 1, 1], [4, 0, 0, 1, 1]])), (None, torch.tensor([[3, 0, 0, 1, 1]]))])
        class_names0 = ['label0', 'label1', 'label2', 'label3', 'label4']
        class_names1 = ['label0', 'label1', 'label02', 'label03', 'label04']
        outputs = Task(Task.Config(use_same_classspace=False, task_type='object_detection')).execute(Task.Inputs(dataset0, dataset1, class_names0, class_names1))
        self.assertEqual(len(outputs.dataset), 6)
        self.assertTrue(torch.equal(outputs.dataset[0][1], torch.tensor([[1, 0, 0, 1, 1]])))
        self.assertTrue(torch.equal(outputs.dataset[3][1], torch.tensor([[1, 0, 0, 1, 1]])))
        self.assertTrue(torch.equal(outputs.dataset[5][1], torch.tensor([[6, 0, 0, 1, 1]])))

    def test_no_second_dataset(self):
        dataset0 = FakeDataset([(None, torch.tensor(1)), (None, torch.tensor(2)), (None, torch.tensor(3))])
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset0, None))
        self.assertEqual(outputs.dataset, dataset0)



"""Model Exponential Moving Average (EMA) aligned with step decay.

This plugin keeps expornential moving average of the model parameter as it is training.

"""
from collections import OrderedDict
import logging
import torch
from .plugin_base import PluginBase


logger = logging.getLogger(__name__)


class ExponentialSmoothingState(object):
    """Exponential moving average of model parameters.
    It maintains a smoothed copy of the model's parameter tensors.

    Args:
        model (torch.nn.Module): Model with parameters whose EMA will be kept.
        decay (float): Decay factor for the exponential moving average.
        device ('cuda', 'cpu', None): Device to store the EMA parameters.
    """
    def __init__(self, model, decay: float = 0.999, device=None):
        self.decay = decay
        if device:
            device = torch.device(device)
        else:
            device = next(model.parameters()).device
        self.device = device

        # Register model parameters
        self.load_shadow_from_model(model)

    @torch.no_grad()
    def load_shadow_from_model(self, model):
        self.shadow = OrderedDict()
        for name, param in model.named_parameters():
            if param.requires_grad:
                self.shadow[name] = param.data.clone().detach().to(self.device)

    @torch.no_grad()
    def step(self, model, steps):
        # Decay parameters more quickly at the beginning to avoid retaining the random initialization
        decay = min(self.decay, (steps + 1.) / (steps + 10))
        for name, param in model.named_parameters():
            if param.requires_grad:
                assert name in self.shadow
                self.shadow[name].copy_((1.0 - decay) * param.detach().to(self.device) + decay * self.shadow[name])


class Plugin(PluginBase):
    def __init__(self, decay: float = 0.999):
        if not (0.0 <= decay <= 1.0):
            raise ValueError(f"Decay must be in the range [0, 1]. decay={decay}")
        self._decay = decay
        self.weight_smoothing_states = None
        self._counter = 0

    def on_train_start(self, trainer, model):
        self._init_ema(model)
        self._counter = 0
        self._weights_updated = False

    def on_optimizer_step_start(self, trainer, model, optimizer):
        self._weights_updated = True
        return optimizer

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        if self._weights_updated:  # Weights might not be updated when gradient accumulation was used.
            self._counter += 1
            self.weight_smoothing_states.step(model, self._counter)
            self._weights_updated = False

    def on_train_end(self, trainer, model):
        if self.weight_smoothing_states:
            self._copy_weights(self.weight_smoothing_states.shadow, list(model.named_parameters()))
            logger.info("Updated the model with the Step Decay EMA parameters.")
        self.weight_smoothing_states = None

    def on_validation_start(self, trainer, model):
        if self.weight_smoothing_states:
            self._model_weights_backup = {n: p.detach().clone() for n, p in model.named_parameters()}
            self._copy_weights(self.weight_smoothing_states.shadow, list(model.named_parameters()))
            logger.debug("Copying the Step Decay EMA parameters to the model.")

    def on_validation_end(self, trainer, model):
        if self.weight_smoothing_states:
            self._copy_weights(self._model_weights_backup, list(model.named_parameters()))
            self._model_weights_backup = None

    def _init_ema(self, model):
        self.weight_smoothing_states = ExponentialSmoothingState(model, self._decay)
        logger.info(f"Model Step Decay EMA is enabled with decay={self._decay}, on {self.weight_smoothing_states.device}.")

    @staticmethod
    def _copy_weights(src_params, dst_params):
        for name, dst_param in dst_params:
            if dst_param.requires_grad:
                dst_param.detach().copy_(src_params[name].to(dst_param.device))



import dataclasses
import logging
import time
import typing
import torch.utils.data
import irisml.core
from irisml.tasks.train.build_dataloader import build_dataloader

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Benchmark dataset loading and preprocessing

    Config:
        batch_size (int): Batch size
        num_workers (int): Number of workers for data loading
        num_batches (int): Number of batches to iterate over. If 0, iterate over all batches.
    """
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset
        transform: typing.Callable

    @dataclasses.dataclass
    class Config:
        batch_size: int
        num_workers: int = 4
        num_batches: int = 0

    @dataclasses.dataclass
    class Outputs:
        time_per_batch: float = 0.0

    def execute(self, inputs):
        dataloader = build_dataloader(inputs.dataset, inputs.transform, batch_size=self.config.batch_size, num_workers=self.config.num_workers, shuffle=False, drop_last=False)

        num_batches = min(self.config.num_batches, len(dataloader)) if self.config.num_batches > 0 else len(dataloader)
        start_time = time.time()
        first_batch_time = 0.0
        for i, _ in enumerate(dataloader):
            if i == 0:
                first_batch_time = time.time() - start_time
            elif i >= num_batches:
                break
        end_time = time.time()

        time_per_batch = (end_time - start_time) / num_batches
        logger.info(f"Took {end_time - start_time} seconds to iterate over {num_batches} batches of size {self.config.batch_size}")
        logger.info(f"First batch: {first_batch_time:.3f} seconds")
        logger.info(f"Time per batch: {time_per_batch:.3f} seconds")

        return self.Outputs(time_per_batch=time_per_batch)



"""Run the model in half precision.

Not recommended in the trianing phase. Use AMP plugin instead.
"""
import logging
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


def _make_half(tensor_or_list):
    if isinstance(tensor_or_list, list):
        return [_make_half(x) for x in tensor_or_list]
    return tensor_or_list.half()


class Plugin(PluginBase):
    def on_train_start(self, trainer, model):
        logger.info("Converting the model to FP16.")
        model.half()

    def on_train_end(self, trainer, model):
        model.float()

    def on_train_batch_start(self, trainer, model, batch, batch_index):
        assert len(batch) == 2
        return _make_half(batch[0]), batch[1]

    def on_validation_batch_start(self, trainer, model, batch, batch_index):
        assert len(batch) == 2
        return _make_half(batch[0]), batch[1]

    def on_prediction_batch_start(self, trainer, model, batch, batch_index):
        assert len(batch) == 2
        return _make_half(batch[0]), batch[1]

    def on_prediction_start(self, trainer, model):
        logger.info("Converting the model to FP16.")
        model.half()

    def on_prediction_end(self, trainer, model):
        model.float()



import unittest
import PIL.Image
import torch
from irisml.tasks.make_classification_dataset_from_predictions import Task


class FakeDataset(torch.utils.data.Dataset):
    def __init__(self, num_images):
        super().__init__()
        self._num_images = num_images

    def __len__(self):
        return self._num_images

    def __getitem__(self, index):
        return PIL.Image.new('RGB', (100, 100)), torch.zeros(1)


class TestMakeClassificationDatasetFromPredictions(unittest.TestCase):
    def test_2d_predictions(self):
        dataset = FakeDataset(10)
        predictions = torch.zeros(10, 3)
        predictions[:, 1] = 1
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset=dataset, predictions=predictions))
        self.assertIsInstance(outputs.dataset, torch.utils.data.Dataset)
        self.assertIsInstance(outputs.dataset[0][1], torch.Tensor)
        self.assertEqual(len(outputs.dataset), 10)
        self.assertEqual([int(outputs.dataset[i][1]) for i in range(10)], [1] * 10)

    def test_1d_predictions(self):
        dataset = FakeDataset(10)
        predictions = torch.ones(10)
        outputs = Task(Task.Config()).execute(Task.Inputs(dataset=dataset, predictions=predictions))
        self.assertIsInstance(outputs.dataset, torch.utils.data.Dataset)
        self.assertIsInstance(outputs.dataset[0][1], torch.Tensor)
        self.assertEqual(len(outputs.dataset), 10)
        self.assertEqual([int(outputs.dataset[i][1]) for i in range(10)], [1] * 10)



import dataclasses
import logging
import pathlib
import typing
import PIL.Image
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Load a simple classification dataset from a directory of images and an index file.

    Config:
        filepath (Path): Path to the index file. The index file should be a text file with one line per image.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        filepath: pathlib.Path

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset
        num_classes: int
        class_names: typing.List[str]

    def execute(self, inputs):
        dataset = SimpleClassificationDataset(self.config.filepath)
        return self.Outputs(dataset=dataset, num_classes=len(dataset.class_names), class_names=dataset.class_names)

    def dry_run(self, inputs):
        return self.execute(inputs)


class SimpleClassificationDataset(torch.utils.data.Dataset):
    def __init__(self, index_filepath):
        super().__init__()
        self._index_filepath = index_filepath
        self._base_dir = index_filepath.parent

        index_lines = index_filepath.read_text().splitlines()
        self._image_indexes = [(path, int(label)) for line in index_lines for path, label in [line.split()]]

        max_class_id = max(label for _, label in self._image_indexes)

        labels_filepath = self._base_dir / 'labels.txt'
        if labels_filepath.exists():
            self._class_names = labels_filepath.read_text().splitlines()
            if max_class_id >= len(self._class_names):
                raise ValueError(f"Number of classes in {labels_filepath} is less than max class id in {index_filepath}")
        else:
            self._class_names = [str(i) for i in range(max_class_id + 1)]

        logger.info(f"Loaded Dataset {index_filepath} with {len(self)} images and {len(self._class_names)} classes")

    def __len__(self):
        return len(self._image_indexes)

    def __getitem__(self, index):
        image_path, label = self._image_indexes[index]
        image = PIL.Image.open(self._base_dir / image_path)
        return image, torch.tensor(label)

    @property
    def class_names(self):
        return self._class_names

    def get_targets(self, index):
        return self._image_indexes[index][1]



import dataclasses
import logging
import math
import irisml.core
import torch

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Convert number of iterations to number of epochs. Min value is 1."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        batch_size: int
        num_iterations: int
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        num_epochs: int

    def execute(self, inputs):
        num_epochs = max(math.ceil(inputs.batch_size * inputs.num_iterations / len(inputs.dataset)), 1)

        return self.Outputs(num_epochs)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
import torch
from irisml.tasks.make_feature_extractor_model import Task


class TestMakeFeatureExtractorModel(unittest.TestCase):
    def test_simple(self):
        class FakeModel(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.a = torch.nn.Conv2d(3, 3, 3)
                self.b = torch.nn.Conv2d(3, 3, 3)
                self.c = torch.nn.Conv2d(3, 3, 3)
                self.d = torch.nn.Sequential(self.a, self.b, self.c)  # Note that the self.a is used twice.

            def forward(self, x):
                a_out = self.a(x)
                return self.d(self.c(self.b(a_out))), a_out

        model = FakeModel()
        inputs = Task.Inputs(model)
        outputs = Task(Task.Config('a')).execute(inputs)

        input_tensor = torch.rand(1, 3, 27, 27)
        _, a_out = model(input_tensor)
        extracted = outputs.model.prediction_step(input_tensor)

        # If a module is used twice the first feature vector will be returned.
        self.assertTrue(torch.equal(a_out, extracted))

        # It supports nested module name.
        outputs = Task(Task.Config('d.0')).execute(inputs)
        extracted = outputs.model.prediction_step(input_tensor)
        self.assertTrue(torch.equal(a_out, extracted))



import unittest
from irisml.tasks.make_fixed_text_dataset import Task


class TestMakeFixedTextDataset(unittest.TestCase):
    def test_simple(self):
        texts = ['str1', 'str2', 'str3']
        outputs = Task(Task.Config()).execute(Task.Inputs(texts))
        dataset = outputs.dataset

        self.assertEqual(len(dataset), 3)
        self.assertEqual(dataset[0], (('str1', []), None))
        self.assertEqual(dataset[1], (('str2', []), None))
        self.assertEqual(dataset[2], (('str3', []), None))



import unittest
import PIL.Image
import torch
import torchvision
from irisml.tasks.make_fixed_prompt_image_transform import Task


class TestMakeFixedPromptImageTransform(unittest.TestCase):
    def test_simple(self):
        image_transform = torchvision.transforms.ToTensor()
        prompt = 'What is in the image?'

        outputs = Task(Task.Config(prompt=prompt)).execute(Task.Inputs(image_transform=image_transform))

        transformed = outputs.transform(PIL.Image.new('RGB', (32, 32)), torch.tensor(3))
        self.assertEqual(len(transformed), 2)
        self.assertEqual(len(transformed[0]), 2)
        self.assertEqual(transformed[0][0], prompt)
        self.assertEqual(len(transformed[0][1]), 1)
        self.assertIsInstance(transformed[0][1][0], torch.Tensor)
        self.assertEqual(transformed[0][1][0].shape, (3, 32, 32))
        self.assertEqual(transformed[1], torch.tensor(3))
